#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#ifndef M_PI
#define M_PI (3.14159265358979323846264338328)
#endif

#include "raycast.h"



struct raycast rc;

void raycast_init(void)
{
	int txx;

	struct intersect ic;

	// first set some initial parameters


	rc.scr_width = 256;
	rc.scr_height = (-1); // to be defined later
	rc.scr_height_max = 192;

	rc.angle_degrees = 70; // 60
	rc.angle_radians = (-666); // to be defined later

	rc.scr_full_y = rc.scr_height_max/8; // rc.scr_height_max/4   // zero is middle

	rc.h = 256.0;

	rc.scr_scroll_steps = 6;

	rc.tx_mx = (-1);
	rc.tx_y_size = (-1);

	rc.bit_y0 = (-1);
	rc.bit_z0 = (-1);

	rc.linemaps = NULL;
	rc.diffmaps = NULL;

	rc.bit_width      = 70; // 80
	rc.bit_height_max = 240;
	rc.bit_height = (-1);
	
	rc.txbyte_size = 5; // 5
	rc.txblist = NULL;
	rc.txbnum = 0;




	// calc remaining parameters

	// size of texture line in memory (2^n)
	txx = rc.bit_width / rc.txbyte_size;
	
	rc.mem_tx_x_size = 1;
	while( rc.mem_tx_x_size < txx )
	{
		rc.mem_tx_x_size <<= 1;
	}

	
	// angle
	rc.angle_radians = (float)(rc.angle_degrees * (M_PI / 180.0));


	// X scaling
	rc.tx_mx = ((double)rc.bit_width)/((double)rc.scr_width);


	// calc initial (low) yz coords of texture plane
	ic=raycast_calc_y_intersect(&rc, ((double)rc.scr_height_max)*0.5 );

	rc.bit_y0 = ic.y;
	rc.bit_z0 = ic.z;



	// calc Y size of texture texel
	ic=raycast_calc_y_intersect(&rc, ((double)rc.scr_height_max)*0.5 - ((double)rc.scr_scroll_steps) );
	//
	rc.tx_y_size = veclen( 0, ic.y, ic.z, 0, rc.bit_y0, rc.bit_z0 );




	//print initial values and some calculated results
	printf("setup values:\n");
	printf("screen size (X*Y): %d x %d\n",rc.scr_width, rc.scr_height_max);
	printf("texture width: %d\n",rc.bit_width);
	printf("texture byte size: %d\n",rc.txbyte_size);
	printf("memory texture X size: %d\n",rc.mem_tx_x_size);
	printf("texture maximum height: %d\n",rc.bit_height_max);
	printf("angle of texture plane (degrees): %f\n",rc.angle_degrees);
	printf("coord of screen Y where full texture line is seen: %d\n",rc.scr_full_y);
	printf("distance to screen plane: %f\n",rc.h);
	printf("initial scroll steps to move single texture texel: %d\n",rc.scr_scroll_steps);

	printf("\ncalculated values:\n");
////	printf("scaling of X texels (at full_y coord): %f\n",rc.tx_mx);
	printf("YZ coords of lower screen line on texture plane: Y=%f, Z=%f\n", rc.bit_y0, rc.bit_z0);
	printf("Y size of single texture texel: %f\n", rc.tx_y_size);

}





struct intersect raycast_calc_y_intersect(struct raycast * rc, double y)
{
	double yfull;
	double t;
	double sina,cosa;
	//
	struct intersect ret;
	//
	yfull = (double)rc->scr_full_y;
	//
	sina = sin(rc->angle_radians);
	cosa = cos(rc->angle_radians);
	//
	t = 1.0 + ( (yfull-y)*sina )/( y*sina + rc->h*cosa );
	//
	ret.y = t * y;
	ret.z = t * rc->h;

	return ret;
}






void raycast_do_maps(void)
{

	int j;

	
	rc.linemaps = (struct linemap *)malloc( rc.scr_height_max * sizeof(struct linemap) );
	if( !rc.linemaps )
	{
		printf("cant alloc rc.linemaps!\n");
		exit(1);
	}

	rc.scr_height = rc.scr_height_max;

	for(j=rc.scr_height_max-1;j>=0;j--)
	{
		double ycur;
		double ydist,ytx;
		double xdist;

		struct intersect ic;

		ycur = ((double)(j-(rc.scr_height_max/2))) + 0.5;

		ic = raycast_calc_y_intersect(&rc, ycur);

		ydist = veclen(0, rc.bit_y0, rc.bit_z0, 0, ic.y, ic.z);

		ytx = ydist / rc.tx_y_size;

		if( ! (ytx>=0 && ytx<((double)rc.bit_height_max)) )
		{
			rc.linemaps[j].tx_y_coord     = (-1);
			rc.linemaps[j].tx_x_scr_width = (-1);

			if( (rc.scr_height_max-1-j)<rc.scr_height )
				rc.scr_height = (rc.scr_height_max-1-j);

			continue;
		}

		rc.linemaps[j].tx_y_coord = ytx;

		// calc X size of texture
		xdist=ic.z; //veclen(0,0,0,0,ic.y,ic.z);
		
		rc.linemaps[j].tx_x_scr_width = (rc.h/xdist)*rc.scr_width;
	
		printf("\nscr_y_coord=%d\n",j);
		mk_x_map(&rc.linemaps[j]);
	}

	
	


	mk_diff_maps();

	
	


}



double veclen(double x0, double y0, double z0, double x1, double y1, double z1)
{
	double dx,dy,dz;

	dx=x1-x0;
	dy=y1-y0;
	dz=z1-z0;

	return sqrt( dx*dx + dy*dy + dz*dz );
}



void mk_x_map(struct linemap * lm)
{
	int int_half_width;
	double half_width,left;
	int i;
	int tx_x;



	half_width = 0.5 * lm->tx_x_scr_width;
	left = ((double)(rc.scr_width/2)) - half_width;

	// 0.5 pix accuracy:
	// 0.5-1.5 - hw=1
	// 1.5-2.5 - hw=2
	// etc.

	int_half_width = (int)floor(half_width+0.5); 


	if( int_half_width > (rc.scr_width/2) )
		int_half_width = (rc.scr_width/2);

	if( int_half_width < 1 )
	{
		printf("half_width is less than 1!\n");
		exit(1);
	}


	lm->xmap.min_scr_pix = ((rc.scr_width/2)-int_half_width);
	lm->xmap.max_scr_pix = ((rc.scr_width/2-1)+int_half_width);

	lm->xmap.right_scr_byte = lm->xmap.max_scr_pix / 8;
	lm->xmap.left_scr_byte  = lm->xmap.min_scr_pix / 8;


	lm->xmap.scr_coords = (int *) malloc( rc.scr_width * sizeof(int) );
	lm->xmap.tx_coords  = (int *) malloc( rc.scr_width * sizeof(int) );
	if( !lm->xmap.scr_coords || !lm->xmap.tx_coords )
	{
		printf("mk_x_map(): cant allocate memory for coords!\n");
		exit(1);
	}

	// fill with placeholders
	for(i=0;i<rc.scr_width;i++)
	{
		lm->xmap.scr_coords[i] = (-1);
		lm->xmap.tx_coords[i]  = (-1);
	}
	
	// then fill with actual values
	for( i=lm->xmap.max_scr_pix; i>=lm->xmap.min_scr_pix; i-- )
	{
		//
		tx_x = (int)floor(rc.bit_width * ( ((double)i) - left + 0.5 ) / lm->tx_x_scr_width );

		
		if(tx_x<0)
		{
			printf("ERROR!\n");
			printf("width=%f, half_width=%f, int_half_width=%d, ",lm->tx_x_scr_width, half_width, int_half_width);
			printf("left=%f\n",left);
			exit(1);
		}

		
		lm->xmap.scr_coords[i] = i;
		lm->xmap.tx_coords[i]  = tx_x;
	}


	//print info
	printf("tx_x_size=%f\n",lm->tx_x_scr_width);
	for(i=rc.scr_width-1;i>=0;i--)
	{
		printf("x=%03d,tx=%02d, ",lm->xmap.scr_coords[i],lm->xmap.tx_coords[i]);
	}
	printf("\n");

	lm->xmap.right_tx_byte = lm->xmap.tx_coords[lm->xmap.max_scr_pix] / rc.txbyte_size;
	lm->xmap.left_tx_byte  = lm->xmap.tx_coords[lm->xmap.min_scr_pix] / rc.txbyte_size;


	mk_txbytes(&lm->xmap);
	

	



	
}


void mk_txbytes(struct xmap * xm)
{
	// make all texture bytes from given xmap (arrays of scr and tx coords)

	int i;
		
	int left_tx,  right_tx;
	int left_scr, right_scr;

	int left_mod, right_mod;

	int * tx  = xm->tx_coords;
	//int * scr = xm->scr_coords;
	
	int bytemap[8];

	struct txbyte tb;
	
	struct txbyteinfo ** cur_tbi_next;
	struct txbyteinfo * cur_tbi = NULL;
	struct txbyteinfo * new_tbi;

	struct txbyte * cur_tb;

	int first_tbi = 1;



	tb.next = NULL;
	tb.bits = bytemap;


	cur_tbi_next = &(xm->right_tbi);
	*(cur_tbi_next) = NULL;
	

	right_tx = rc.scr_width-1;


	// find starting tx byte
	while( tx[right_tx]<0 )
	{
		right_tx--;

		if( right_tx<0 )
		{
			printf("mk_txbytes(): found no valid texture coords!\n");
			exit(1);
		}
	}
	
	while( right_tx>=0 && tx[right_tx]>=0 )
	{
		// select tx byte of size rc.rxbyte_size
		left_tx = right_tx;
		while(  left_tx>=0 && tx[left_tx]>=0 && (tx[right_tx] - (tx[right_tx] % rc.txbyte_size))<=tx[left_tx]  )
		{
			left_tx--;
		}
		left_tx++;

//printf("selected left_tx: [%d]=%d, right_tx: [%d]=%d\n",left_tx,tx[left_tx],right_tx,tx[right_tx]);
	
		// extract and generate textured bytes
		right_scr = right_tx;
		while( right_scr >= left_tx )
		{
			left_scr = right_scr&(~7L);
			left_scr = (left_scr<left_tx) ? left_tx : left_scr;

//printf("selected left_scr=%d, right_scr=%d\n",left_scr,right_scr);

			for(i=0;i<8;i++)
				bytemap[i]=(-1);
		
		
//printf("generating bytemap:\n");
			for(i=right_scr;i>=left_scr;i--)
			{
				bytemap[i&7] = tx[i] % rc.txbyte_size;
//printf("[%d]=%d ",i,tx[i]);
			}
//printf("\n");
//for(i=7;i>=0;i--)
//printf("[%d]=%d, ",i,bytemap[i]);			
//printf("\n");
			
			if( !(cur_tb=find_same_txbyte(&tb)) )
			{
				cur_tb=add_txbyte(&tb);
//printf("unique bytemap added!\n");
			}
//printf("\n");			


			// make new txbyteinfo
			cur_tbi = (struct txbyteinfo *) malloc( sizeof(struct txbyteinfo) );
			if( !cur_tbi )
			{
				printf("mk_txbytes(): cant alloc txbyteinfo!\n");
				exit(1);
			}

			*(cur_tbi_next) = cur_tbi;
			cur_tbi->next = NULL;
			cur_tbi_next = &(cur_tbi->next);

			cur_tbi->txtable = cur_tb;

			cur_tbi->scr_x_byte = left_scr / 8;
			
			cur_tbi->tx_x_byte = tx[left_tx] / rc.txbyte_size;


			left_mod  = left_scr&7;
			right_mod = right_scr&7;


			if( left_mod==0 && right_mod==7 )
			{
				cur_tbi->byte_info = TBI_FULL;
			}
			else if( right_mod==7 )
			{
				cur_tbi->byte_info = TBI_START;
			}
			else if( left_mod==0 )
			{
				cur_tbi->byte_info = TBI_LAST;
			}
			else
			{
				cur_tbi->byte_info = TBI_MID;
			}


			if( first_tbi && cur_tbi->byte_info == TBI_MID )
			{
				cur_tbi->byte_info = TBI_START;
			}

			if( first_tbi && cur_tbi->byte_info == TBI_LAST )
			{
				cur_tbi->byte_info = TBI_FULL;
			}




			first_tbi=0;



			right_scr = (right_scr&(~7L)) - 1;
		}


		// set next right_tx
		right_tx = left_tx-1;
	}


	if( cur_tbi->byte_info == TBI_START )
	{
		cur_tbi->byte_info = TBI_FULL;
	}
	else if( cur_tbi->byte_info == TBI_MID )
	{
		cur_tbi->byte_info = TBI_LAST;
	}




	// we have built right tbi chain, now make left tbi chain from it
	cur_tbi = xm->right_tbi;
	xm->left_tbi = NULL;
	
	while( cur_tbi )
	{
		new_tbi = (struct txbyteinfo *) malloc( sizeof(struct txbyteinfo) );
		if( !new_tbi )
		{
			printf("mk_txbytes(): cant alloc txbyteinfo!\n");
			exit(1);
		}
		
		memcpy( new_tbi, cur_tbi, sizeof(struct txbyteinfo) );
		
		new_tbi->next = xm->left_tbi;
		xm->left_tbi = new_tbi;
		
		switch( new_tbi->byte_info )
		{
		case TBI_START:
			new_tbi->byte_info = TBI_LAST;
		break;
		
		case TBI_LAST:
			new_tbi->byte_info = TBI_START;
		break;
		}
	
		cur_tbi = cur_tbi->next;
	}
	
	
	
	


// report tbis
/*printf("right_scr_byte=%d, right_tx_byte=%d\n",xm->right_scr_byte,xm->right_tx_byte);
cur_tbi = xm->right_tbi;
while( cur_tbi )
{
	printf("right: tbi_scrbyte=%d, tby_txbyte=%d, tbi_txtablenum=%d, TBI_",
		cur_tbi->scr_x_byte,
		cur_tbi->tx_x_byte,
		cur_tbi->txtable->num );
	
	switch(cur_tbi->byte_info)
	{
	case TBI_FULL:  printf("FULL\n");  break;
	case TBI_START: printf("START\n"); break;
	case TBI_LAST:  printf("LAST\n");  break;
	case TBI_MID:   printf("MID\n");   break;
	}

	cur_tbi = cur_tbi->next;
}*/
printf(" left_scr_byte=%d,  left_tx_byte=%d\n",xm->left_scr_byte,xm->left_tx_byte);
cur_tbi = xm->left_tbi;
while( cur_tbi )
{
	printf("left: tbi_scrbyte=%d, tby_txbyte=%d, tbi_txtablenum=%d, TBI_",
		cur_tbi->scr_x_byte,
		cur_tbi->tx_x_byte,
		cur_tbi->txtable->num );
	
	switch(cur_tbi->byte_info)
	{
	case TBI_FULL:  printf("FULL\n");  break;
	case TBI_START: printf("START\n"); break;
	case TBI_LAST:  printf("LAST\n");  break;
	case TBI_MID:   printf("MID\n");   break;
	}

	cur_tbi = cur_tbi->next;
}







}








void mk_diff_maps(void)
{
	int i,j;
	int i2;

	double tx_y1_add, tx_y2_add;

	int tx_y1, tx_y2;

	struct diffmap * dm;




	rc.diffmaps = (struct diffmap **)malloc( rc.scr_scroll_steps * sizeof(struct diffmap *) );
	if( !rc.diffmaps )
	{
		printf("mk_diff_maps(): cant get mem for diffmaps!\n");
		exit(1);
	}



	// all scroll_steps loop
	for(i=rc.scr_scroll_steps-1;i>=0;i--)
	{
		i2 = i+1;

		tx_y1_add = ((double)i ) / ((double)rc.scr_scroll_steps);
		tx_y2_add = ((double)i2) / ((double)rc.scr_scroll_steps);

		if( i2 >= rc.scr_scroll_steps )
		{
			i2 -= rc.scr_scroll_steps;
		}


		rc.diffmaps[i]=NULL;


		// loop through all Y coords to determine changing lines
		for(j=(rc.scr_height_max-1);j>=(rc.scr_height_max-rc.scr_height);j--)
		{
			tx_y1 = (int)floor( tx_y1_add + rc.linemaps[j].tx_y_coord );
			tx_y2 = (int)floor( tx_y2_add + rc.linemaps[j].tx_y_coord );

			if( tx_y1 != tx_y2 )
			{
				dm = (struct diffmap *)malloc( sizeof(struct diffmap) );
				if( !dm )
				{
					printf("mk_diff_maps(): cant get memory for struct diffmap!\n");
					exit(1);
				}


				dm->next = rc.diffmaps[i];
				rc.diffmaps[i] = dm;

				dm->line = &rc.linemaps[j];
				
				dm->y_tx_coord  = tx_y2;
				dm->y_scr_coord = j;
			}
		}
	}




// print diffmap logs
for(i=0;i<rc.scr_scroll_steps;i++)
{
printf("\ndiffmap %d:\n",i);
dm =  rc.diffmaps[i];
while(dm)
{
printf("scr_y=%d, tx_y=%d\n",dm->y_scr_coord,dm->y_tx_coord);
dm=dm->next;
}
}
printf("\n");



}












struct txbyte * find_same_txbyte(struct txbyte * tb)
{
	struct txbyte * curr;
	int i;
	int found;
	
	curr = rc.txblist;
	
	while(curr)
	{
		found = 1;
		
		for(i=0;i<8;i++)
		{
			if( curr->bits[i] != tb->bits[i] )
			{
				found = 0;
				break;
			}
		}

		if( found )
			return curr;
		
		curr = curr->next;
	}
	
	return NULL;
}

struct txbyte * add_txbyte(struct txbyte * tb)
{
	struct txbyte * fresh;
	
	fresh = (struct txbyte *) malloc( sizeof(struct txbyte) );
	if( !fresh )
	{
		printf("add_txbyte(): cant allocate mem!\n");
		exit(1);
	}

	fresh->bits = (int *) malloc( 8 * sizeof(int) );
	if( !fresh->bits )
	{
		printf("add_txbyte(): cant allocate mem!\n");
		exit(1);
	}
	
	// memcpy: dst,src,size
	memcpy( fresh->bits, tb->bits, 8*sizeof(int) );
	
	fresh->next = rc.txblist;
	rc.txblist = fresh;
	fresh->num = rc.txbnum;
	
	rc.txbnum++;

	return fresh;
}


int num_txbytes(void)
{
	return rc.txbnum;
}
