#ifndef _RAYCAST_H_
#define _RAYCAST_H_



struct raycast
{
	int bit_width; // width of texture in bits (==texels)
	int bit_height_max; // height of texture in bits (==texels) <=256
	int bit_height; // real height, if less than _max, or _max

	int scr_width; // width of screen in pixels
	int scr_height; // (same) -- not real height, but used for tmapping
	int scr_height_max; // maximum available

	float angle_degrees;
	float angle_radians;

	int scr_full_y; // Y of screen coords where texture line is mapped at full screen width

	int scr_scroll_steps; // how many scroll steps at nearest line

	double tx_mx;
	double tx_y_size;

	double bit_z0, bit_y0; // beginning of texture plane (corresponds to lower screen line)

	
	double h; // how far is screen from origin

	
	int txbyte_size;

	struct txbyte * txblist;
	int txbnum;

	int mem_tx_x_size; // X size of texture in memory, usually is 2^n
	
	
	struct linemap * linemaps; // all Y lines mapped here

	struct diffmap ** diffmaps; // differential maps (array of ptrs to arrays)

};



struct intersect
{
	double y;
	double z;
};

struct txbyte
{
	int * bits;
	
	int num;
	
	struct txbyte * next;
};



struct txbyteinfo
{
	int scr_x_byte;
	int tx_x_byte;

#define TBI_FULL  (1)
#define TBI_START (2)
#define TBI_MID   (4)
#define TBI_LAST  (8)

	int byte_info;


	struct txbyte * txtable;

	struct txbyteinfo * next;
};


struct xmap
{
	int min_scr_pix;
	int max_scr_pix;

	int * scr_coords;
	int * tx_coords;

	int right_scr_byte;
	int right_tx_byte;

	int left_scr_byte;
	int left_tx_byte;

	

	struct txbyteinfo * right_tbi; // TBIs *from* right
	
	struct txbyteinfo * left_tbi; // TBIs *from* left
};


struct linemap
{
	double tx_y_coord;
	double tx_x_scr_width; // width of all texture plane on screen plane in screen pixels

	struct xmap xmap;
	
};



struct diffmap
{
	struct linemap * line;

	int y_tx_coord;
	int y_scr_coord;

	struct diffmap * next;
};









void raycast_init(void);

struct intersect raycast_calc_y_intersect(struct raycast * rc, double y);

void raycast_do_maps(void);

double veclen(double x0, double y0, double z0, double x1, double y1, double z1);

void mk_x_map(struct linemap * lm);


void mk_txbytes(struct xmap * xm);


void mk_diff_maps(void);




struct txbyte * find_same_txbyte(struct txbyte * tb);
struct txbyte * add_txbyte(struct txbyte * tb);
int num_txbytes(void);



#endif // _RAYCAST_H_

